# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
Pre-configured chopper and detector parameters for Dream.
"""

from ...chopper import Chopper
from ...detector import Detector
from ..common import make_beamline

dream_high_flux = {
    "choppers": {
        "PSC1": {
            "frequency": 14.0,
            "phase": 286.0 - 180.0,
            "distance": 6.145,
            "centers": [0, 72, 86.4, 115.2, 172.8, 273.6, 288.0, 302.4],
            "widths": [2.46, 3.02, 3.27, 3.27, 5.02, 3.93, 3.93, 2.46],
            "direction": "anti-clockwise",
        },
        "PSC2": {
            "frequency": 14.0,
            "phase": 236.0,
            "distance": 6.155,
            "centers": [0, 28.8, 57.6, 144.0, 158.4, 216.0, 259.2, 316.8],
            "widths": [2.46, 3.60, 3.60, 3.23, 3.27, 3.77, 3.94, 2.62],
            "direction": "clockwise",
        },
        "OC": {
            "frequency": 14.0,
            "phase": 297.0 - 180.0 - 90.0,
            "distance": 6.174,
            "centers": [0.0],
            "widths": [27.6],
            "direction": "anti-clockwise",
        },
        "BC": {
            "frequency": 112.0,
            "phase": 240.0 - 180.0,
            "distance": 9.78,
            "centers": [0.0, 180.0],
            "widths": [73.75, 73.75],
            "direction": "anti-clockwise",
        },
        "T0": {
            "frequency": 28.0,
            "phase": 280.0 - 180.0,
            "distance": 13.05,
            "centers": [0.0],
            "widths": [314.9],
            "direction": "anti-clockwise",
        },
    },
    "detectors": {
        "mantle": {"distance": 77.675},
        "end-cap": {"distance": 77.675},
        "high-resolution": {"distance": 79.05},
        "sans": {"distance": 79.05},
    },
}


def dream(
    high_flux=False, high_resolution=False
) -> dict[str, list[Chopper] | list[Detector]]:
    if high_flux and high_resolution:
        raise ValueError("Select either high flux or high resolution, not both.")
    if high_flux:
        return make_beamline(dream_high_flux)
    if high_resolution:
        # TODO: Add high-resolution configuration
        raise NotImplementedError("High-resolution configuration not yet implemented.")
    raise ValueError("Either high_flux or high_resolution must be True.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import scipp as sc

from ..chopper import AntiClockwise, Chopper, Clockwise
from ..detector import Detector


def _array_or_none(container, key):
    return (
        sc.array(dims=["cutout"], values=container[key], unit="deg")
        if key in container
        else None
    )


def make_beamline(instrument) -> dict[str, list[Chopper] | list[Detector]]:
    choppers = [
        Chopper(
            frequency=ch["frequency"] * sc.Unit("Hz"),
            direction={"clockwise": Clockwise, "anti-clockwise": AntiClockwise}[
                ch["direction"]
            ],
            open=_array_or_none(ch, "open"),
            close=_array_or_none(ch, "close"),
            centers=_array_or_none(ch, "centers"),
            widths=_array_or_none(ch, "widths"),
            phase=ch["phase"] * sc.Unit("deg"),
            distance=ch["distance"] * sc.Unit("m"),
            name=key,
        )
        for key, ch in instrument["choppers"].items()
    ]
    detectors = [
        Detector(distance=det["distance"] * sc.Unit("m"), name=key)
        for key, det in instrument["detectors"].items()
    ]
    return {"choppers": choppers, "detectors": detectors}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from itertools import chain

import scipp as sc

from .chopper import Chopper
from .detector import Detector
from .result import Result
from .source import Source

ComponentType = Chopper | Detector


def _input_to_dict(
    obj: None | list[ComponentType] | tuple[ComponentType, ...] | ComponentType,
    kind: type,
):
    if isinstance(obj, list | tuple):
        out = {}
        for item in obj:
            new = _input_to_dict(item, kind=kind)
            for key in new.keys():
                if key in out:
                    raise ValueError(f"More than one component named '{key}' found.")
            out.update(new)
        return out
    elif isinstance(obj, kind):
        return {obj.name: obj}
    elif obj is None:
        return {}
    else:
        raise TypeError(
            "Invalid input type. Must be a Chopper or a Detector, "
            "or a list/tuple of Choppers or Detectors."
        )


class Model:
    """
    A class that represents a neutron instrument.
    It is defined by a list of choppers, a list of detectors, and a source.

    Parameters
    ----------
    choppers:
        A list of choppers.
    detectors:
        A list of detectors.
    source:
        A source of neutrons.
    """

    def __init__(
        self,
        source: Source,
        choppers: Chopper | list[Chopper] | tuple[Chopper, ...] | None = None,
        detectors: Detector | list[Detector] | tuple[Detector, ...] | None = None,
    ):
        self.choppers = _input_to_dict(choppers, kind=Chopper)
        self.detectors = _input_to_dict(detectors, kind=Detector)
        self.source = source

    def add(self, component):
        """
        Add a component to the instrument.
        Component names must be unique across choppers and detectors.

        Parameters
        ----------
        component:
            A chopper or detector.
        """
        if component.name in chain(self.choppers, self.detectors):
            raise KeyError(
                f"Component with name {component.name} already exists. "
                "If you wish to replace/update an existing component, use "
                "``model.choppers['name'] = new_chopper`` or "
                "``model.detectors['name'] = new_detector``."
            )
        if isinstance(component, Chopper):
            self.choppers[component.name] = component
        elif isinstance(component, Detector):
            self.detectors[component.name] = component
        else:
            raise TypeError(
                f"Cannot add component of type {type(component)} to the model."
            )

    def remove(self, name: str):
        """
        Remove a component.

        Parameters
        ----------
        name:
            The name of the component to remove.
        """
        if name in self.choppers:
            del self.choppers[name]
        elif name in self.detectors:
            del self.detectors[name]
        else:
            raise KeyError(f"No component with name {name} was found.")

    def __iter__(self):
        return chain(self.choppers, self.detectors)

    def __getitem__(self, name) -> Chopper | Detector:
        if name not in self:
            raise KeyError(f"No component with name {name} was found.")
        return self.choppers[name] if name in self.choppers else self.detectors[name]

    def __delitem__(self, name):
        self.remove(name)

    def run(self):
        """
        Run the simulation.
        """
        components = sorted(
            chain(self.choppers.values(), self.detectors.values()),
            key=lambda c: c.distance.value,
        )

        birth_time = self.source.data.coords['birth_time']
        speed = self.source.data.coords['speed']
        initial_mask = sc.ones(sizes=birth_time.sizes, unit=None, dtype=bool)

        result_choppers = {}
        result_detectors = {}
        time_limit = (
            birth_time + (components[-1].distance / speed).to(unit=birth_time.unit)
        ).max()
        for c in components:
            container = result_detectors if isinstance(c, Detector) else result_choppers
            container[c.name] = c.as_dict()
            container[c.name]['data'] = self.source.data.copy(deep=False)
            t = birth_time + (c.distance / speed).to(unit=birth_time.unit, copy=False)
            container[c.name]['data'].coords['toa'] = t
            container[c.name]['data'].coords['eto'] = t % (
                1 / self.source.frequency
            ).to(unit=t.unit, copy=False)
            container[c.name]['data'].coords['distance'] = c.distance
            # TODO: remove 'tof' coordinate once deprecation period is over
            container[c.name]['data'].coords['tof'] = t
            if isinstance(c, Detector):
                container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
                continue
            m = sc.zeros(sizes=t.sizes, unit=None, dtype=bool)
            to, tc = c.open_close_times(time_limit=time_limit)
            container[c.name].update({'open_times': to, 'close_times': tc})
            for i in range(len(to)):
                m |= (t > to[i]) & (t < tc[i])
            combined = initial_mask & m
            container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
            container[c.name]['data'].masks['blocked_by_me'] = ~m & initial_mask
            initial_mask = combined

        return Result(
            source=self.source, choppers=result_choppers, detectors=result_detectors
        )

    def __repr__(self) -> str:
        out = f"Model:\n  Source: {self.source}\n  Choppers:\n"
        for name, ch in self.choppers.items():
            out += f"    {name}: {ch}\n"
        out += "  Detectors:\n"
        for name, det in self.detectors.items():
            out += f"    {name}: {det}\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from .dream import dream
from .odin import odin
from .pulse_profile import pulse

__all__ = ["dream", "odin", "pulse"]
